@@ -5,6 +5,15 @@
 #' @useDynLib lwgeom
 NULL
 
+.onLoad = function(libname, pkgname) {
+	# SET PROJ.db path:
+	if (file.exists(system.file("proj/nad.lst", package = "sf")[1])) { # we are on Windows:
+		prj = system.file("proj", package = "sf")[1]
+		CPL_set_data_dir(prj) # if existing, uses C API to set path
+		CPL_use_proj4_init_rules(1L)
+	}
+}
+
 .onAttach = function(libname, pkgname) {
 	esv = lwgeom_extSoftVersion()
 	m = paste0("Linking to liblwgeom ", esv["lwgeom"],
@@ -17,6 +26,8 @@ NULL
 		warning(paste("GEOS versions differ: lwgeom has", esv["GEOS"], "sf has", sf["GEOS"]))
 	if (sf["proj.4"] != esv["proj.4"])
 		warning(paste("PROJ versions differ: lwgeom has", esv["proj.4"], "sf has", sf["proj.4"]))
+
+	
 }
 
 #' Provide the external dependencies versions of the libraries linked to sf
